#include "lexer.h"
#include "panic.h"
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void error_construct(Error* error, Pos pos, String message)
{
    *error = (Error) {
        .pos = pos,
        .message = message,
    };
}

void error_destroy(Error* error) { string_destroy(&error->message); }

void error_collector_construct(ErrorCollector* collector)
{
    const size_t errors_start_capacity = 64;
    *collector = (ErrorCollector) {
        .errors = malloc(errors_start_capacity * sizeof(Error)),
        .errors_length = 0,
        .errors_capacity = errors_start_capacity,
    };
}

void error_collector_destroy(ErrorCollector* collector)
{
    for (size_t i = 0; i < collector->errors_length; ++i) {
        error_destroy(&collector->errors[i]);
    }
    free(collector->errors);
}

void error_collector_add(ErrorCollector* collector, Error error)
{
    if (collector->errors_length >= collector->errors_capacity) {
        Error* new_buffer = realloc(
            collector->errors, collector->errors_capacity * sizeof(Error)
        );
        ASSERT(new_buffer);
        collector->errors = new_buffer;
    }
    collector->errors[collector->errors_length] = error;
    collector->errors_length += 1;
}

const char* token_type_value(TokenType type)
{
    switch (type) {
        case TokenTypeEof:
            return "Eof";
        case TokenTypeError:
            return "error";
        case TokenTypeId:
            return "Id";
        case TokenTypeInt:
            return "Int";
        case TokenTypeChar:
            return "Char";
        case TokenTypeString:
            return "String";
        case TokenTypeLParen:
            return "(";
        case TokenTypeRParen:
            return ")";
        case TokenTypeLBrace:
            return "{";
        case TokenTypeRBrace:
            return "}";
        case TokenTypeLBracket:
            return "[";
        case TokenTypeRBracket:
            return "]";
    }
}

#define LEXER_ADD_ERROR(LEXER, POS, ...)                                       \
    {                                                                          \
        String error_message;                                                  \
        string_construct(&error_message);                                      \
        string_append_format(&error_message, __VA_ARGS__);                     \
        Error error;                                                           \
        error_construct(&error, (POS), error_message);                         \
        error_collector_add((LEXER)->errors, error);                           \
    }

Lexer lexer_create(CharReader* reader, ErrorCollector* errors)
{
    return (Lexer) {
        .reader = reader,
        .errors = errors,
        .current = reader->next(reader),
        .pos = { 0, 1, 1 },
    };
}

Token lexer_next(Lexer* lexer)
{
    Pos pos = lexer->pos;
    if (lexer_done(lexer)) {
        return lexer_token(lexer, TokenTypeEof, pos);
    }
    switch (lexer->current) {
        case ' ':
        case '\t':
        case '\r':
        case '\n':
            return lexer_skip_whitespace(lexer);
        case '"':
            return lexer_lex_string(lexer);
        case '(':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeLParen, pos)
            );
        case ')':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeRParen, pos)
            );
        case '{':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeLBrace, pos)
            );
        case '}':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeRBrace, pos)
            );
        case '[':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeLBracket, pos)
            );
        case ']':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeRBracket, pos)
            );
        case '/':
            return lexer_lex_slash(lexer);
        case '#':
            return lexer_lex_hashtag(lexer);
        case '0' ... '9':
            return lexer_lex_int(lexer);
        case '+':
        case '-':
        case '*':
        case '<':
        case '>':
        case '=':
        case '!':
        case 'a' ... 'z':
        case 'A' ... 'Z':
            return lexer_lex_id(lexer);
        default:
            LEXER_ADD_ERROR(
                lexer, pos, "unsupported character `%c`", lexer->current
            );
            return (lexer_step(lexer), lexer_token(lexer, TokenTypeError, pos));
    }
}

Token lexer_lex_id(Lexer* lexer)
{
    Pos pos = lexer->pos;
    lexer_step(lexer);
    while (true) {
        switch (lexer->current) {
            case '+':
            case '-':
            case '*':
            case '<':
            case '>':
            case '=':
            case '!':
            case '0' ... '9':
            case 'a' ... 'z':
            case 'A' ... 'Z':
                lexer_step(lexer);
                break;
            default:
                goto break_loop;
        }
    }
break_loop:
    return lexer_token(lexer, TokenTypeId, pos);
}

Token lexer_lex_int(Lexer* lexer)
{
    Pos pos = lexer->pos;
    lexer_step(lexer);
    while (true) {
        switch (lexer->current) {
            case '0' ... '9':
                lexer_step(lexer);
                break;
            default:
                goto break_loop;
        }
    }
break_loop:
    return lexer_token(lexer, TokenTypeInt, pos);
}

Token lexer_lex_char(Lexer* lexer)
{
    Pos pos = lexer->pos;
    lexer_step(lexer);
    if (lexer_done(lexer)) {
        LEXER_ADD_ERROR(
            lexer, pos, "malformed character literal, got unexpected Eof"
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    if (lexer->current == '\'') {
        LEXER_ADD_ERROR(
            lexer, pos, "malformed character literal, got unexpected `'`"
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    if (lexer->current == '\\') {
        lexer_step(lexer);
        if (lexer_done(lexer)) {
            LEXER_ADD_ERROR(
                lexer, pos, "malformed character literal, got unexpected Eof"
            );
            return lexer_token(lexer, TokenTypeError, pos);
        }
        lexer_step(lexer);
    }
    if (lexer_done(lexer)) {
        LEXER_ADD_ERROR(
            lexer, pos, "malformed character literal, expected `'`, got Eof"
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    if (lexer->current != '\'') {
        LEXER_ADD_ERROR(
            lexer,
            pos,
            "malformed character literal, expected `'`, got `%c`",
            lexer->current
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    lexer_step(lexer);
    return lexer_token(lexer, TokenTypeChar, pos);
}

Token lexer_lex_string(Lexer* lexer)
{
    Pos pos = lexer->pos;
    lexer_step(lexer);
    while (!lexer_done(lexer)) {
        char maybe_escape_char = lexer->current;
        lexer_step(lexer);
        if (maybe_escape_char == '\\') {
            if (lexer_done(lexer)) {
                LEXER_ADD_ERROR(
                    lexer,
                    pos,
                    "malformed string literal escape sequence, got unexpected "
                    "Eof"
                );
                return lexer_token(lexer, TokenTypeError, pos);
            }
            lexer_step(lexer);
        }
    }
    if (lexer_done(lexer)) {
        LEXER_ADD_ERROR(
            lexer, pos, "malformed string literal, expected `\"`, got Eof"
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    if (lexer->current != '\"') {
        LEXER_ADD_ERROR(
            lexer,
            pos,
            "malformed string literal, expected `\"`, got `%c`",
            lexer->current
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    lexer_step(lexer);
    return lexer_token(lexer, TokenTypeString, pos);
}

Token lexer_lex_slash(Lexer* lexer)
{
    Pos pos = lexer->pos;
    lexer_step(lexer);
    if (lexer->current == '*') {
        lexer_step(lexer);
        int depth = 1;
        char last = '\0';
        while (!lexer_done(lexer) && depth > 0) {
            if (last == '*' && lexer->current == '/') {
                depth -= 1;
            } else if (last == '/' && lexer->current == '*') {
                depth += 1;
            }
            last = lexer->current;
            lexer_step(lexer);
        }
        if (depth > 0) {
            String error_message;
            string_from_cstr(
                &error_message,
                "malformed multiline comment, expected `*/` before end"
            );
            Error error;
            error_construct(&error, pos, error_message);
            error_collector_add(lexer->errors, error);
            return lexer_token(lexer, TokenTypeError, pos);
        }
        return lexer_next(lexer);
    }
    if (lexer->current == '/') {
        lexer_step(lexer);
        while (!lexer_done(lexer) && lexer->current != '\n') {
            lexer_step(lexer);
        }
        return lexer_next(lexer);
    }
    return lexer_token(lexer, TokenTypeId, pos);
}

Token lexer_lex_hashtag(Lexer* lexer)
{
    lexer_step(lexer);
    while (!lexer_done(lexer) && lexer->current != '\n') {
        lexer_step(lexer);
    }
    return lexer_next(lexer);
}

Token lexer_skip_whitespace(Lexer* lexer)
{
    lexer_step(lexer);
    while (true) {
        switch (lexer->current) {
            case ' ':
            case '\t':
            case '\r':
            case '\n':
                lexer_step(lexer);
                break;
            default:
                goto break_loop;
        }
    }
break_loop:
    return lexer_next(lexer);
}

Token lexer_token(const Lexer* lexer, TokenType type, Pos pos)
{
    return (Token) {
        .type = type,
        .pos = pos,
        .length = lexer->pos.index - pos.index,
    };
}

void lexer_step(Lexer* lexer)
{
    if (lexer->current == '\n') {
        lexer->pos.line += 1;
        lexer->pos.col = 1;
    } else if (!lexer_done(lexer)) {
        lexer->pos.col += 1;
    }
    lexer->pos.index += 1;
    lexer->current = lexer->reader->next(lexer->reader);
}

bool lexer_done(const Lexer* lexer) { return lexer->current == '\0'; }
