#include "vec.h"
#include "panic.h"
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

bool string_slice_equal(StringSlice slice, const char* data)
{
    if (strlen(data) != slice.length) {
        return false;
    }
    return strncmp(data, slice.data, slice.length) == 0;
}

void string_construct(String* string)
{
    *string = (String) {
        .data = NULL,
        .length = 0,
        .capacity = 0,
    };
}

void string_destroy(String* string)
{
    if (string->data) {
        free(string->data);
    }
}

const size_t string_starting_alloc_size = 8;

void string_from_cstr(String* string, const char* value)
{
    string_construct(string);
    string_append_cstr(string, value);
}

void string_from_slice(String* string, StringSlice slice)
{
    string_construct(string);
    for (size_t i = 0; i < slice.length; ++i) {
        string_append_char(string, slice.data[i]);
    }
}

void string_append(String* string, const String* value)
{
    string_append_cstr(string, value->data);
}

void string_append_char(String* string, char value)
{
    if (string->length + 1 >= string->capacity) {
        if (string->capacity == 0) {
            string->capacity = string_starting_alloc_size;
        } else {
            string->capacity *= 2;
        }
        if (string->data == NULL) {
            string->data = malloc(string->capacity * sizeof(char));
            ASSERT(string->data);
        } else {
            char* new_buffer
                = realloc(string->data, string->capacity * sizeof(char));
            ASSERT(new_buffer);
            string->data = new_buffer;
        }
    }
    string->data[string->length] = value;
    string->length += 1;
    string->data[string->length] = '\0';
}

void string_append_cstr(String* string, const char* value)
{
    size_t value_length = strlen(value);

    if (string->length + value_length + 1 > string->capacity) {
        if (string->capacity == 0) {
            string->capacity = string_starting_alloc_size;
        }
        while (string->length + value_length + 1 > string->capacity) {
            string->capacity *= 2;
        }
        char* new_buffer
            = realloc(string->data, string->capacity * sizeof(char));
        ASSERT(new_buffer);
        string->data = new_buffer;
    }

    memcpy(&string->data[string->length], value, value_length);
    string->length += value_length;
    string->data[string->length] = '\0';
}

void string_append_format(String* string, const char* format, ...)
{
    va_list varargs;
    va_start(varargs, format);

    size_t format_length = (size_t)vsnprintf(NULL, 0, format, varargs);

    va_end(varargs);

    va_start(varargs, format);

    if (string->length + format_length + 1 > string->capacity) {
        if (string->capacity == 0) {
            string->capacity = string_starting_alloc_size;
        } else {
            string->capacity *= 2;
        }
        while (string->length + format_length + 1 > string->capacity) {
            string->capacity *= 2;
        }
        char* new_buffer
            = realloc(string->data, string->capacity * sizeof(char));
        ASSERT(new_buffer);
        string->data = new_buffer;
    }

    size_t written = (size_t)vsnprintf(
        &string->data[string->length],
        string->capacity - string->length,
        format,
        varargs
    );
    string->length += written;
    string->data[string->length] = '\0';

    va_end(varargs);
}

bool string_equal(const String* string, const char* value)
{
    return strncmp(value, string->data, string->length + 1);
}
